/*********************************************
 * vim:sw=8:ts=8:si:et
 * To use the above modeline in vim you must have "set modeline" in your .vimrc
 * Author: Guido Socher
 * Copyright: GPL V2
 * See http://www.gnu.org/licenses/gpl.html
 *
 * Ethernet remote switch (4 independent switches)
 * UDP and HTTP interface 
 *
 * Chip type           : Atmega88 or Atmega168 or Atmega328 or Atmega644 with ENC28J60
 *********************************************/
#include <avr/io.h>
#include <stdlib.h>
#include <string.h>
#include <avr/pgmspace.h>
#include "ip_arp_udp_tcp.h"
#include "websrv_help_functions.h"
#include "enc28j60.h"
#include "timeout.h"
#include "net.h"

// This software is a web server only. 
//
// please modify the following two lines. mac and ip have to be unique
// in your local area network. You can not have the same numbers in
// two devices:
static uint8_t mymac[6] = {0x54,0x55,0x58,0x10,0x00,0x29};
// how did I get the mac addr? Translate the first 3 numbers into ascii is: TUX
static uint8_t myip[4] = {10,0,0,29};
//static uint8_t myip[4] = {192,168,255,100};
// listen port for tcp/www:
static uint16_t mywwwport=80;
//
// Max amount of switches (0..3) to display on the webpage:
#define MAX_RELAY 4
// the positions on PORTD where the relays are connected:
static uint8_t relay2port[MAX_RELAY] = {4,5,6,7}; // PD4, PD5, PD6, PD7
// uncomment to use the port expander via 74hc595
#define USE_74595 1
// global string buffer
#define STR_BUFFER_SIZE 24
static char gStrbuf[STR_BUFFER_SIZE+1];
static uint16_t gPlen;
#define BUFFER_SIZE 650
static uint8_t buf[BUFFER_SIZE+1];
// name of this switch (title on the main webpage)
static char label[21]="IO-web server"; // must not be longer than 20 char
//
// password (only a-z,0-9,_ characters):
static char password[9]="secret"; // must not be longer than 9 char
//
#ifdef USE_74595
static uint8_t e74595val=0;
#endif
// set output to VCC, red LED off
#define LEDOFF PORTB|=(1<<PORTB1)
// set output to GND, red LED on
#define LEDON PORTB&=~(1<<PORTB1)
// to test the state of the LED
#define LEDISOFF PORTB&(1<<PORTB1)
// 
uint8_t verify_password(char *str)
{
        // the first characters of the received string are
        // a simple password/cookie:
        if (strncmp(password,str,strlen(password))==0){
                return(1);
        }
        return(0);
}

uint16_t http200ok(void)
{
        return(fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\nPragma: no-cache\r\n\r\n")));
}

void usage(void){
        gPlen=fill_tcp_data_p(buf,http200ok(),PSTR("<p>Usage: http://host_or_ip/password</p>\n"));
}

#ifdef USE_74595
// SER pin 14:
#define S74595_0 PORTC&=~(1<<PORTC5)
#define S74595_1 PORTC|=(1<<PORTC5)
// RCLK pin 12:
#define S74595_RCLKDOWN PORTC&=~(1<<PORTC4)
#define S74595_RCLKUP PORTC|=(1<<PORTC4)
// SRCLK pin 11:
#define S74595_CLOCKDOWN PORTC&=~(1<<PORTC3)
#define S74595_CLOCKUP PORTC|=(1<<PORTC3)
// Inline assembly, nop = do nothing for a clock cycle.
//#define nop()  asm volatile("nop\n\t" "nop\n\t"::)
#define nop()  asm volatile("nop\n\t"::)
void set74595(uint8_t val)
{
      uint8_t i=8;
      while(i){
              i--;
              S74595_CLOCKDOWN;
              if (val & (1<<i)){
                      S74595_1;
              }else{
                      S74595_0;
              }
              S74595_CLOCKUP;
              nop();
      }
      S74595_CLOCKDOWN;
      S74595_RCLKUP;
      nop();
      S74595_RCLKDOWN;
      e74595val=val;
}
void init74595(void)
{
        DDRC|= (1<<DDC4); // enable as output line
        S74595_RCLKDOWN;
        DDRC|= (1<<DDC5); // enable as output line
        S74595_0;
        DDRC|= (1<<DDC3); // enable as output line
        S74595_CLOCKDOWN;
        e74595val=0; // variable where we keep the current state
        set74595(0);
}
#endif

void print_page_show_switch(uint8_t val){
        uint16_t pl;
        pl=http200ok();
        if (val){
                gPlen=fill_tcp_data(buf,pl,"1");
        }else{
                gPlen=fill_tcp_data(buf,pl,"0");
        }
}

// answer HTTP/1.0 301 Moved Permanently\r\nLocation: .....\r\n\r\n
// to redirect
// type =0  : http://tuxgraphics.org/c.ico    favicon.ico file
// type =1  : /password/
uint16_t moved_perm(uint8_t *buf,uint8_t type)
{
        uint16_t plen;
        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 301 Moved Permanently\r\nLocation: "));
        if (type==1){
                plen=fill_tcp_data_p(buf,plen,PSTR("/"));
                plen=fill_tcp_data(buf,plen,password);
                plen=fill_tcp_data_p(buf,plen,PSTR("/"));
        }else{
                plen=fill_tcp_data_p(buf,plen,PSTR("http://tuxgraphics.org/ico/sw.ico"));
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("\r\n\r\nContent-Type: text/html\r\n\r\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h1>301 Moved Permanently</h1>\n"));
        return(plen);
}

// main webpage
uint16_t print_webpage_main(void)
{
        uint16_t pl;
        uint8_t switchnum=0; 
        pl=http200ok();
        pl=fill_tcp_data_p(buf,pl,PSTR("<a href=./>[refresh]</a>"));
        pl=fill_tcp_data_p(buf,pl,PSTR("<h2>"));
        pl=fill_tcp_data(buf,pl,label);
        pl=fill_tcp_data_p(buf,pl,PSTR("</h2>\n<pre>"));

        while(switchnum<MAX_RELAY){ 
                if (PORTD & (1<<relay2port[switchnum])){
                        pl=fill_tcp_data_p(buf,pl,PSTR("ON "));
                }else{
                        pl=fill_tcp_data_p(buf,pl,PSTR("OFF"));
                }
                pl=fill_tcp_data_p(buf,pl,PSTR(" <a href=s"));
                itoa(switchnum,gStrbuf,10); // convert integer to string
                pl=fill_tcp_data(buf,pl,gStrbuf); // switch number in ascii
                pl=fill_tcp_data(buf,pl,"-");
                if (PORTD & (1<<relay2port[switchnum])){
                        pl=fill_tcp_data(buf,pl,"0");
                }else{
                        pl=fill_tcp_data(buf,pl,"1");
                }
                pl=fill_tcp_data_p(buf,pl,PSTR(">s"));
                itoa(switchnum,gStrbuf,10); // convert integer to string
                pl=fill_tcp_data(buf,pl,gStrbuf); // switch number in ascii
                pl=fill_tcp_data_p(buf,pl,PSTR("</a>"));
                pl=fill_tcp_data(buf,pl,"\n");
                switchnum++;
        }
#ifdef USE_74595
        pl=fill_tcp_data(buf,pl,"\n");
        switchnum=0;
        while(switchnum<8){ // port extended switch 0..7 via 74hc595 
                if (e74595val & (1<<switchnum)){
                        pl=fill_tcp_data_p(buf,pl,PSTR("ON "));
                }else{
                        pl=fill_tcp_data_p(buf,pl,PSTR("OFF"));
                }
                pl=fill_tcp_data_p(buf,pl,PSTR(" <a href=e"));
                itoa(switchnum,gStrbuf,10); // convert integer to string
                pl=fill_tcp_data(buf,pl,gStrbuf);
                pl=fill_tcp_data(buf,pl,"-");
                if (e74595val & (1<<switchnum)){
                        pl=fill_tcp_data(buf,pl,"0");
                }else{
                        pl=fill_tcp_data(buf,pl,"1");
                }
                pl=fill_tcp_data_p(buf,pl,PSTR(">e"));
                itoa(switchnum,gStrbuf,10); // convert integer to string
                pl=fill_tcp_data(buf,pl,gStrbuf); // switch number in ascii
                pl=fill_tcp_data_p(buf,pl,PSTR("</a>"));
                pl=fill_tcp_data(buf,pl,"\n");
                switchnum++;
        }
#endif
        pl=fill_tcp_data_p(buf,pl,PSTR("</pre><hr>"));
        return(pl);
}

// analyse the url given
// return values: -1 invalid password
//                -2 redirect to same url but ending in slash
//                10 web page buffer already filled with data
//                0 print main web page
//
//                The string passed to this function will look like this:
//                /password/s0=1 HTTP/1.....
//                /password/s0=0 HTTP/1.....
//                /password HTTP/1.....
int8_t analyse_get_url(char *str)
{
        uint8_t loop=15;
        int8_t i;
        // the first slash:
        if (*str == '/'){
                str++;
        }else{
                return(-1);
        }
        if (str[0]==' '){ // end of url, main page
                usage();
                return(10);
        }
        if (strncmp("favicon.ico",str,11)==0){
                // favicon:
                gPlen=moved_perm(buf,0);
                return(10);
        }
        // the password:
        if(verify_password(str)==0){
                return(-1);
        }
        // move forward to the first space or '/'
        while(loop){
                if(*str==' '){
                        // end of url and no slash after password:
                        return(-2);
                }
                if(*str=='/'){
                        // end of password
                        loop=0;
                        continue;
                }
                str++;
                loop--; // do not loop too long
        }
        // str is now something like password/sw0-1 or just end of url
        if (strncmp("/s",str,2)==0){
                str+=2;
                i=*str-'0'; // char to integer
                if (i>=0 && i<MAX_RELAY){ // positions 0 to MAX_RELAY-1 in relay2port
                        str++;
                        if (*str!='-'){
                                return(-1);
                        }
                        str++;
                        if (*str=='0'){
                                PORTD &= ~(1<<relay2port[i]);// transistor off
                                return(0);
                        }
                        if (*str=='1'){
                                PORTD|= (1<<relay2port[i]);// transistor on
                                return(0);
                        }
                        // switch value is not 1 and not 0, e.g "x":
                        print_page_show_switch(PORTD & (1<<relay2port[i]));
                        return(10);
                }
                return(-1);
        }
#ifdef USE_74595
        // str is now something like password/es0-1 or just end of url
        if (strncmp("/e",str,2)==0){
                str+=2;
                i=*str-'0'; // char to integer
                if (i>=0 && i<=7){ // switch 0 to 7
                        str++;
                        if (*str!='-'){
                                return(-1);
                        }
                        str++;
                        if (*str=='0'){
                                set74595(e74595val&~(1<<i));// off
                                return(0);
                        }
                        if (*str=='1'){
                                set74595(e74595val|(1<<i));// on
                                return(0);
                        }
                        // switch value is not 1 and not 0, e.g "x":
                        print_page_show_switch(e74595val & (1<<i));
                        return(10);
                }
        }
#endif
        // just a slash after the password:
        if (strncmp("/ ",str,2)==0){
                return(0);
        }
        return(-1);
}


int main(void){
        uint16_t dat_p;
        int8_t cmd;
        
#ifdef USE_74595
        init74595(); // it is best to do 74hc595 init right after power on
#endif

        // set the clock speed to "no pre-scaler" (8MHz with internal osc or 
        // full external speed)
        // set the clock prescaler. First write CLKPCE to enable setting 
        // of clock the next four instructions.
        // Note that the CKDIV8 Fuse determines the initial
        // value of the CKKPS bits.
        CLKPR=(1<<CLKPCE); // change enable
        CLKPR=0; // "no pre-scaler"
        _delay_loop_1(0); // 60us

        /*initialize enc28j60*/
        enc28j60Init(mymac);
        enc28j60clkout(2); // change clkout from 6.25MHz to 12.5MHz
        _delay_loop_1(0); // 60us
        
        /* Magjack leds configuration, see enc28j60 datasheet, page 11 */
        // LEDB=yellow LEDA=green
        //
        // 0x476 is PHLCON LEDA=links status, LEDB=receive/transmit
        // enc28j60PhyWrite(PHLCON,0b0000 0100 0111 01 10);
        enc28j60PhyWrite(PHLCON,0x476);
        
        //DDRB|= (1<<DDB1); // enable PB1, LED as output 
        // the transistor on PD7 and others as output:
        DDRD|= (1<<DDD7);
        DDRD|= (1<<DDD6);
        DDRD|= (1<<DDD5);
        DDRD|= (1<<DDD4);
        PORTD &= ~(1<<PORTD7);// transistor off
        PORTD &= ~(1<<PORTD6);// transistor off
        PORTD &= ~(1<<PORTD5);// transistor off
        PORTD &= ~(1<<PORTD4);// transistor off
        
        //init the web server ethernet/ip layer:
        init_ip_arp_udp_tcp(mymac,myip,mywwwport);

        while(1){

                // handle ping and wait for a tcp packet
                gPlen=enc28j60PacketReceive(BUFFER_SIZE, buf);
                buf[BUFFER_SIZE]='\0';
                dat_p=packetloop_icmp_tcp(buf,gPlen);

                if(dat_p==0){
                        // not tcp
                        continue;
                }
                if (strncmp("GET ",(char *)&(buf[dat_p]),4)!=0){
                        // head, post and other methods:
                        //
                        // for possible status codes see:
                        // http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
                        gPlen=http200ok();
                        gPlen=fill_tcp_data_p(buf,gPlen,PSTR("<h1>200 OK</h1>"));
                        goto SENDTCP;
                }
                cmd=analyse_get_url((char *)&(buf[dat_p+4]));
                // for possible status codes see:
                // http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
                if (cmd==-1){
                        gPlen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 401 Unauthorized\r\nContent-Type: text/html\r\n\r\n<h1>401 Unauthorized</h1>"));
                        goto SENDTCP;
                }
                if (cmd==10){
                        goto SENDTCP;
                }
                if (cmd==-2){
                        // redirect to the right base url (e.g add a trailing slash):
                        gPlen=moved_perm(buf,1);
                        goto SENDTCP;
                }
                // just display the status:
                gPlen=print_webpage_main();
                //
SENDTCP:
                www_server_reply(buf,gPlen); // send data
                continue;

                // tcp port www end
        }
        return (0);
}
